__doc__ = 'Strategy objects to process HiperDom directives'

import utils
from xml.dom.Node import Node
from DomWrapper import AlreadyRendered, toxml

# Namespace URI for HiperDom attributes
HDOM_NAMESPACE = 'uri:hiperdom'

class _strategy:
    def __init__ (self, context, REQUEST, template, kw={}):
        self.r = REQUEST
        self.kw = kw
        self.context = context
        self.template = template

class assetHunter (_strategy):
    def __init__ (self, parent, preserveDirectives=0):
        self.preserveDirectives = preserveDirectives
        self.parent = parent
        _strategy.__init__(self, parent.context, parent.r, parent.template,
            parent.kw)

    def do_id (self, element, outDoc, specialData):
        """Process the data and add the results to specialData with
        the name specified in the hdom:id directive
        """
        name = element.getAttributeNS (HDOM_NAMESPACE, 'id')
        wkcopy = outDoc.importNode(element, 1)
        # We don't want to remove the hdom:id, probably because we're running
        # from inside a textExpander
        if not self.preserveDirectives:
            # remove the hdom:id attribute
            wkcopy.removeAttributeNS (HDOM_NAMESPACE, 'id')
        specialData[name] = toxml(self.template._processElement(wkcopy, 
            self.parent, outDoc))
        return wkcopy

class textExpander (_strategy):
    def __init__ (self, context, REQUEST, template, kw={}, 
            preserveDirectives=0):
        self.preserveDirectives = preserveDirectives
        _strategy.__init__(self, context, REQUEST, template, kw)
        
    def do_text (self, element, outDoc, specialData):
        """Replace the contents (childNodes) of the element by the
        result of evaluating the identifier passed as the value of
        the hdom:text directive.
        But first, hunt for any assets in the subtree to be discarded
        (assets are elements with an hdom:id attribute); the assets
        are added to the namespace when the identifier is evaluated.
        Finally, here we don't want to strip the Hdom directives
        unless explicitly told so
        """
        name = element.getAttributeNS (HDOM_NAMESPACE, 'text')
        wkcopy = outDoc.importNode(element, 0)
        assets = self.kw.copy()
        if not self.preserveDirectives:
            wkcopy.removeAttributeNS (HDOM_NAMESPACE, 'text')
        self.template._processElement(element, 
            assetHunter(self, self.preserveDirectives), outDoc, assets)

        value = utils.evaluate(name, self.context, self.r, assets)
        
        wkcopy.appendChild(AlreadyRendered(outDoc, str(value)))
        return wkcopy

class fullHdom (textExpander):
    def do_dummy (self, element, outDoc, specialData):
        """Discard this element"""
        return None

    def do_condition (self, element, outDoc, specialData):
        """Evaluate the name given as the directive's value, then leave the
        contents untouched if the name evaluates to true and discard the
        element if it evaluates to false.
        """
        name = element.getAttributeNS (HDOM_NAMESPACE, 'condition')

        if utils.evaluate(name, self.context, self.r, self.kw):
            wkcopy = outDoc.importNode(element, 1)
            wkcopy.removeAttributeNS (HDOM_NAMESPACE, 'condition')
            return wkcopy
        else:
            return None

    def do_loop (self, element, outDoc, specialData):
        """Evaluate the name given as the directive's value, then process
        the contents of the element once for every item of the result.
        It is an error for the result not to be a sequence object.
        """
        name = element.getAttributeNS (HDOM_NAMESPACE, 'loop')
        fragment = outDoc.createDocumentFragment()
        namespace = self.kw.copy()

        for item in utils.evaluate(name, self.context, self.r, self.kw):
            namespace['sequence_item'] = item
            if hasattr (item, '__of__'):
                newStrategy = fullHdom(item.__of__(self.context), self.r, 
                    self.template, namespace)
            else:
                # FIXME: not sure
                newStrategy = fullHdom(self.context, self.r, self.template, 
                    namespace)
            
            wkcopy = outDoc.importNode(element, 1)
            wkcopy.removeAttributeNS (HDOM_NAMESPACE, 'loop')
            child = self.template._processElement(wkcopy, newStrategy, outDoc)
            fragment.appendChild(child)
        
        return fragment

    def do_method (self, element, outDoc, specialData):
        """Replace the whole element by a new element resulting from
        evaluating the identifier passed as the value of the hdom:method
        directive.
        The method will be called with a special parameter: element_data,
        which is the xml.dom.Element object.        
        It may return a dictionary with some special conventions (keys are
        attributes, a special key '<text>' is the content), or a string

        or a new xml.dom Element object.
        """
        name = element.getAttributeNS (HDOM_NAMESPACE, 'method')
        wkcopy = outDoc.importNode(element, 1)
        assets = self.kw.copy()
        wkcopy.removeAttributeNS (HDOM_NAMESPACE, 'method')
        wkcopy.__allow_access_to_unprotected_subobjects = 1
        assets['element_data'] = wkcopy
        result = utils.evaluate(name, self.context, self.r, assets)
        if result is None or hasattr(result, 'nodeType'):
            return result
        elif type(result) == type(''):
            return AlreadyRendered(outDoc, result)
        elif type (result) == type ({}):
            newtext = ''
            if result.has_key ('<text>'):
                newtext = result['<text>']
                del result['<text>']
            for name in wkcopy.attributes.keys():
                wkcopy.removeAttribute(name)
            for name, value in result.items():
                wkcopy.setAttributeNS('', name, value)
            for child in wkcopy.childNodes[:]:
                wkcopy.removeChild(child)
            if newtext:
                wkcopy.appendChild(AlreadyRendered(outDoc, newtext))
            return wkcopy
        else:
            raise TypeError, 'invalid object returned from hdom:method'

