__doc__ = """
HiperDom Template objects
HiperDom Templates are XML page templates - meaning, they can generate XML
documents if they're processed in a Zope (for example) context. The primary
use for them is generating xHTML webpages.

HiperDom works by recognizing special XML attributes (called Directives),
marked with a "hdom:" XML namespace. For example,
<title hdom:text="docTitle">foo</title> will replace "foo" by the value
resulting from evaluating the variable, property or function "docTitle".
"""
__version__ = '0.4.0'

import Globals, Acquisition, AccessControl.Role, OFS.SimpleItem, OFS.PropertyManager, DateTime
import string
import HiperDomStrategies, DomWrapper, utils
from xml.dom.Node import Node

_hdom_deprecated_attrs = ()
_hdom_prototype = {'HTML': 0, 
        'text': "<?xml version='1.0' encoding='UTF-8'?>\n<EmptyTemplate xmlns:hdom='uri:hiperdom'/>",
        '_dom': None, 'title': 'HiperDom Template', 'content_type': '',
    }

HDOM_NAMESPACE = HiperDomStrategies.HDOM_NAMESPACE
SIMPLE_NODES = (Node.ATTRIBUTE_NODE, Node.TEXT_NODE, Node.TEXT_NODE,
    Node.CDATA_SECTION_NODE, Node.ENTITY_NODE, Node.PROCESSING_INSTRUCTION_NODE,
    Node.COMMENT_NODE, Node, Node.NOTATION_NODE)

class HiperDomTemplate (
    OFS.SimpleItem.Item,
    Globals.Persistent,
    Acquisition.Implicit,
    AccessControl.Role.RoleManager,
    OFS.PropertyManager.PropertyManager
    ):
    """HiperDom Template class
    __call__ and __str__ render the template;
    index_html renders the template and sets the proper content-type on RESPONSE;
    manage_getTemplate returns the source, and sets content-type;
    text_expand processes only hdom:text directives, leaving them in place and
        respecting hdom:id assets
    The template object is "meaningful" in the namespace, or in other words,
        it's parallel to a DTML Document, not a DTML Method
    """

    #################################################
    ### Zope class attributes

    name = 'HiperDomTemplate'
    id = 'HiperDomTemplate'
    meta_type = 'HiperDom Template'
    icon = 'misc_/HiperDom/HiperDomTemplate.gif'
    # Tell DocumentTemplate that we accept its calling conventions...
    isDocTemp = 1
    # ...then tell HiperDom that we prefer ours
    isXmlTemp = 1

    _properties = (
            {'id': 'HTML', 'type': 'boolean', 'mode': 'w'},
            {'id': 'title', 'type': 'string', 'mode': 'w'},
            {'id': 'content_type', 'type': 'string', 'mode': 'w'},
        )

    manage_options = (
            {'label':'Edit', 'action':'manage_main',
             'help': ('HiperDom', 'TemplateEdit.stx')},
            {'label':'Expand', 'action':'manage_expand',
             'help': ('HiperDom', 'TemplateExpand.stx')},
            {'label':'View', 'action':''},
            {'label':'Properties', 'action':'manage_propertiesForm',
             'help': ('OFSP', 'Properties.stx')},
            {'label':'Security', 'action':'manage_access',
             'help': ('OFSP', 'Security.stx')},
            {'label':'Ownership', 'action':'manage_owner',
             'help': ('OFSP', 'Ownership.stx')},
        )


    __ac_permissions__ = (
            ('View management screens',
             ('manage_tabs', 'manage_main', 'manage_expand',
              'manage_getTemplate', 'text_expand', 'document_src')),
            ('Change Permissions',
             ('manage_access',)),
            ('Change HiperDom Templates',
             ('manage_edit', 'PUT')),
            ('Manage Properties',
             ('manage_addProperty', 'manage_editProperties',
             'manage_delProperties', 'manage_changeProperties',)),
            ('View',
             ('', 'index_html', '__call__',)),
            ('FTP access',
             ('manage_FTPstat','manage_FTPget','manage_FTPlist')),
            ('Know the secrets of life',
             ('manage_dict',)),
            )


    #################################################
    ## Instance constructors

    def __init__ (self, id):
        "Initialize the class instance"
        self.id = id
        for name, value in _hdom_prototype.items ():
            if not hasattr (self, name):
                setattr (self, name, value)
        if not self._dom:
            self._updateDom()

    def __setstate__ (self, state):
        "Initialize when loading from the ZODB"
        Globals.Persistent.__setstate__ (self, state)
        for name, value in _hdom_prototype.items ():
            if not hasattr (self, name):
                setattr (self, name, value)
        for name in _hdom_deprecated_attrs:
            if hasattr (self, name):
                delattr (self, name)
        if not self._dom or not self._dom.documentElement:
            self._updateDom()


    #################################################
    ## This does the "Narower/Wider/Taller/Shorter"
    ## manage_main textarea magic
    ## (shamelessly copied from OFS.DTMLMethod)

    _size_changes = {
        'Bigger': (5,5),
        'Smaller': (-5,-5),
        'Narrower': (0,-5),
        'Wider': (0,5),
        'Taller': (5,0),
        'Shorter': (-5,0),
    }

    def _er (self, text, submit, dtpref_cols, dtpref_rows, REQUEST):
        dr, dc = self._size_changes[submit]
        
        rows = max (1, string.atoi (dtpref_rows) + dr)
        cols = max (40, string.atoi (dtpref_cols) + dc)
        e = (DateTime.DateTime ('GMT') + 365).rfc822()
        resp = REQUEST['RESPONSE']
        resp.setCookie ('dtpref_rows', str (rows), path='/', expires=e)
        resp.setCookie ('dtpref_cols', str (cols), path='/', expires=e)
        return self.manage_main (self, REQUEST, text=text, dtpref_cols=cols, dtpref_rows=rows)


    #################################################
    ## Management interface

    manage_main = Globals.HTMLFile ('HiperDomTemplateEdit', globals())
    manage_expand = Globals.HTMLFile ('HiperDomTemplateExpand', globals())

    def manage_edit (self, text, submit='Change', dtpref_cols='50',
                    dtpref_rows='20', REQUEST=None):
        "Edit the template."
        if self._size_changes.has_key (submit):
             return self._er (text, submit, dtpref_cols, dtpref_rows, REQUEST)
        
        self.text = text
        self._updateDom()

        if REQUEST is not None:
            return self.manage_main (self, REQUEST,
                manage_tabs_message = 'Template changed.')


    #################################################
    ## Externally accessible, but not from the manage
    ## interface
    
    def index_html (self, REQUEST, **kw):
        "Render the template into a webpage"
        result = apply (self.__str__, (self, REQUEST), kw)
        self._setContentType (REQUEST)
        return result

    def manage_getTemplate (self, REQUEST=None):
        """Return the template
        It is rendered from the dom tree instead of
        just self.text, so that we return a normalized
        version (or XMLized if it was HTML)"""
        if REQUEST:
            self._setContentType (REQUEST)
        return DomWrapper.toxml(self._dom)
    
    # this alias makes WebDAV happy
    # (it allows you to use the "source link property" to
    # get the source instead of the rendered template, when
    # you're retrieving the document for edition via DAV)
    document_src = manage_getTemplate

    # This alias is to make FTP get possible.
    manage_FTPget = manage_getTemplate

    def PUT (self, REQUEST, RESPONSE):
        "Change the template from an HTTP PUT"
        self.dav__init (REQUEST, RESPONSE)
        self.text = REQUEST.get ('BODY', '')
        self._updateDom()
        RESPONSE.setStatus(204)
        return RESPONSE


    #################################################
    ## Externally accessible from Python/DTML only
    
    def __str__ (self, context=None, REQUEST=None, **kw):
        "Render the template"
        if context is None:
            context = self
        if REQUEST is None:
            if hasattr (context, 'REQUEST'):
                REQUEST = context.REQUEST
            else:
                REQUEST = {}
        return DomWrapper.toxml(self._processElement(self._dom, 
            HiperDomStrategies.fullHdom(context, REQUEST, self, kw)))

    def __call__ (self, context=None, REQUEST=None, **kw):
        "Render the template given a context (?)"
        # DocumentTemplate calls us with (None, context)
        if context is None and hasattr (REQUEST, 'has_key') and REQUEST.has_key ('REQUEST'):
            context = REQUEST
            REQUEST = context['REQUEST']
        # FIXME: how should I use the context? Where does Acquisition fit
        # into this?
        return apply (self.__str__, (context, REQUEST), kw)

    def text_expand (self, REQUEST=None, **kw):
        """Process only hdom:text directives, leaving them in place and
        respecting hdom:id assets
        """
        if not REQUEST:
            REQUEST=self.REQUEST
        return DomWrapper.toxml(self._processElement(self._dom, 
            HiperDomStrategies.textExpander(self, REQUEST, self, kw, 
                preserveDirectives=1)))


    #################################################
    ## Methods for internal consumption
    
    def _updateDom (self):
        "Update the internal DOM tree from the text"
        self._dom = DomWrapper.todom(self.text)
    
    def _setContentType (self, REQUEST):
        "Set the content-type"
        # FIXME: we want more robustness, like checking if the
        # first element in the DOM is an <html> and if we have a
        # Content-type property
        if self.content_type:
            ct = self.content_type
            if string.find(ct, 'charset') == -1:
                ct = '%s;charset=UTF-8' % ct
            REQUEST.RESPONSE.setHeader('Content-type', ct)
            return

        if self._dom.doctype.name == 'html':
            REQUEST.RESPONSE.setHeader('Content-type', 
                'text/html;charset=UTF-8')

        return # default content-type handling

    default_content_type = 'text/xml;charset=UTF-8'
        
    def _processElement (self, element, strategy, outDoc=None, specialData={}):
        """hdom:process one element (and its children) using this strategy
        returns the new, processed Node object (or an DomWrapper.AlreadyRendered
        pseudo-Node if the strategy returned a string)
        specialData may be used to "fish" for data in the tree; for
        example, it is used to gather assets marked with hdom:id
        before gobbling an hdom:text element
        """

        # Special case: the document element should be copied to form the basis
        # of the output DOM. It is then subsequently used to import all rendered
        # elements.
        if element.nodeType == element.DOCUMENT_NODE:
            if element.doctype is not None: 
                docType = element.doctype.cloneNode(1)
            else:
                docType = None 
            outDoc = element.implementation.createDocument(None, None, docType)
            
            # Clone any non-document children of the template DOM; they can't be
            # elements and thus are not affected by the rendering process.
            for child in element.childNodes: 
                if child.nodeType == child.DOCUMENT_TYPE_NODE: continue
                if child == element.documentElement: continue
                outDoc.appendChild(outDoc.importNode(child, 0))

            if element.documentElement:
                newDocEl = self._processElement(element.documentElement,
                    strategy, outDoc, specialData)
                if newDocEl:
                    outDoc.appendChild(newDocEl)

            return outDoc
        
        assert outDoc != None, 'No output document'
        
        # Skip nodes that don't have attributes anyway
        if element.nodeType in SIMPLE_NODES:
            return outDoc.importNode(element, 1)
        
        # handle hdom: attributes
        attrs = element.attributes
        if attrs:
            for key in filter(lambda k: k[0] == HDOM_NAMESPACE, attrs.keys()):
                if hasattr(strategy, 'do_' + key[1]):
                    wkcopy = getattr(strategy, 'do_' + key[1])(element,
                        outDoc, specialData)
                    if type (wkcopy) == type (''):
                        return DomWrapper.AlreadyRendered(outDoc, wkcopy)
                    else:
                        # We don't want to process the childNodes of already
                        # processed elements. Processed elements take care of
                        # that themselves. 
                        return wkcopy

        # we don't want to modify the upstream element object
        wkcopy = outDoc.importNode(element, 0)

        # process the children
        for child in element.childNodes:
            newChild = self._processElement(child, strategy, outDoc, 
                specialData)
            if newChild:
                wkcopy.appendChild(newChild)
        return wkcopy

    #################################################
    ## Debugging stuff, in the "what not even God
    ## should do" category
    def manage_dict (self, REQUEST=None):
        "Debugging method - meaning is obvious"
        return `self.__dict__`

manage_addHiperDomTemplateForm = Globals.HTMLFile ('HiperDomTemplateAdd', globals ())

def manage_addHiperDomTemplate (self, id, REQUEST=None):
    "Add an HiperDomTemplate to the current ObjectManager"
    obj = HiperDomTemplate (id)
    assert obj, "Could not create template!"
    obj.title = REQUEST['title']
    self._setObject (id, obj)
    if REQUEST is not None:
        return self.manage_main (self, REQUEST)
